﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO.Ports;
using Modbus;

namespace Modbus
{
    public partial class MainWindow : Form, IDataReceiveListener
    {
        private ModbusMaster modbusMaster;
        private ModbusSlave modbusSlave;
        private SerialConnector serialConnector;
        private bool bMaster = true;

        public MainWindow()
        {
            InitializeComponent();

            LoadForm();

            serialConnector = new SerialConnector(this);
            modbusMaster = new ModbusMasterASCII();
            modbusSlave = new ModbusSlaveASCII();
        }

        private void LoadForm()
        {
            // Ports
            string[] ports = SerialPort.GetPortNames();
            foreach (string port in ports)
            {
                cbPort.Items.Add(port);
            }

            // Baudrate
            string[] baudrates = { "230400", "115200", "57600", "38400", "19200", "9600" };
            foreach (string b in baudrates)
            {
                cbBaudrate.Items.Add(b);
            }

            // Word length            
            string[] wordLengths = { "5", "6", "7", "8" };
            foreach (string wl in wordLengths)
            {
                cbWordLength.Items.Add(wl);
            }

            // Stop bit
            string[] stopBits = { "0", "1", "1.5", "2" };
            foreach (string sb in stopBits)
            {
                cbStopBit.Items.Add(sb);
            }

            // Parity bit
            string[] parityBits = { "Brak", "Nieparzyste", "Parzyste", "Znak", "Pusty" };
            foreach (string pb in parityBits)
            {
                cbParityBit.Items.Add(pb);
            }

            // Work mode
            string[] workMode = { "ASCII", "RTU" };
            foreach (string wm in workMode)
            {
                cbMode.Items.Add(wm);
            }
            cbMode.SelectedIndex = 0;

            // Master Functions
            string[] masterFns = { "Fn01 - Read Coils",
                                   "Fn02 - Read Discrete Inputs",
                                   "Fn03 - Read Holding Registers", 
                                   "Fn04 - Read Input Registers", 
                                   "Fn05 - Write Single Coil",
                                   "Fn06 - Write Sinle Register" };
            foreach (string fn in masterFns)
            {
                cbMasterFns.Items.Add(fn);
            }

            cbBaudrate.SelectedItem = "115200";
            cbWordLength.SelectedItem = "8";
            cbStopBit.SelectedItem = "1";
            cbParityBit.SelectedItem = "Brak";

            cbSlaveID.SelectedIndex = 0;
            modbusSlave.SlaveID = (uint)cbSlaveID.SelectedIndex + 1;
        }

        private void bClose_Click(object sender, EventArgs e)
        {
            Close();
        }

        private StopBits getStopBits()
        {
            //string[] stopBits = { "0", "1", "1.5", "2" };
            switch(cbStopBit.SelectedIndex)
            {
                case 0:
                    return StopBits.None;
                case 1:
                    return StopBits.One;
                case 2:
                    return StopBits.OnePointFive;
                case 3:
                    return StopBits.Two;
                default:
                    return StopBits.None;
            }
        }

        private Parity getParity()
        {
             //string[] parityBits = { "Brak", "Nieparzyste", "Parzyste", "Znak", "Pusty" };
            switch (cbParityBit.SelectedIndex)
            {
                case 0:
                    return Parity.None;
                case 1:
                    return Parity.Odd;
                case 2:
                    return Parity.Even;
                case 3:
                    return Parity.Mark;
                case 4:
                    return Parity.Space;
                default:
                    return Parity.None;
            }
        }

        private void bStart_Click(object sender, EventArgs e)
        {
            if (cbPort.SelectedIndex < 0)
            {
                MessageBox.Show("Wybierz port COM", "Blad");
                return;
            }

            bool bResult = serialConnector.Open((string)cbPort.SelectedItem,
                                        int.Parse((string)cbBaudrate.SelectedItem),
                                        int.Parse((string)cbWordLength.SelectedItem),
                                        getStopBits(),
                                        getParity());

            if (bResult)
            {
                cbPort.Enabled = false;
                cbBaudrate.Enabled = false;
                cbWordLength.Enabled = false;
                cbStopBit.Enabled = false;
                cbParityBit.Enabled = false;
                bSend.Enabled = true;
            }
            
            PrintLog(serialConnector.GetLastStatus());
        }

        private void bStop_Click(object sender, EventArgs e)
        {
            bool bResult = serialConnector.Close();

            if (bResult)
            {
                cbPort.Enabled = true;
                cbBaudrate.Enabled = true;
                cbWordLength.Enabled = true;
                cbStopBit.Enabled = true;
                cbParityBit.Enabled = true;
                bSend.Enabled = false;
            }
            PrintLog(serialConnector.GetLastStatus());
        }

        #region IReceiveHandler Members
        private delegate void ProcessRequestDelegate(byte[] data);

        public void DataReceivedEvent(byte[] data)
        {
            this.Invoke(new ProcessRequestDelegate(this.ProcessRequest), new object[] { data });
        }

        #endregion

        private void ProcessRequest(byte[] data)
        {
            if (bMaster)
            {
                PrintFrame("Response: ", data);
            }
            else
            {
                PrintFrame("Request: ", data);
                byte[] response = modbusSlave.HandleRequest(data);
                if (modbusSlave.LastStatus != null)
                {
                    PrintLog(modbusSlave.LastStatus);
                }
                if (response != null)
                {
                    serialConnector.SendData(response);
                    PrintFrame("Response: ", response);
                }
            }
        }

        private void PrintLog(string text)
        {
            lbLog.Items.Add(text);
            lbLog.TopIndex = lbLog.Items.Count - 1;
        }

        private void PrintFrame(string text, byte[] data)
        {
            PrintLog(text + new ASCIIEncoding().GetString(data));
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            serialConnector.StopReceveing();
        }

        private void bSend_Click(object sender, EventArgs e)
        {
            byte[] request = null;

            if (CheckAddressRegisters())
            {
                if (cbMasterFns.SelectedIndex > -1)
                {
                    try
                    {
                        request = modbusMaster.GetFnRequest(tbSlaveAddress.Text,
                                                            cbMasterFns.SelectedIndex + 1,
                                                            tbRegOff.Text,
                                                            tbValue.Text);
                    }
                    catch (Exception)
                    {
                        // styknie
                        MessageBox.Show("Nieprawidłowe wywołanie funkcji");
                    }
                }
                else
                {
                    MessageBox.Show("Wybierz funkcję");
                }
            }
            else
            {
                MessageBox.Show("Nieprawidłowe wywołanie funkcji");
            }

            if (request != null)
            {
                PrintFrame("Request: ", request);
                serialConnector.SendData(request);
            }
        }

        private bool CheckAddressRegisters()
        {
            return !tbSlaveAddress.Text.Equals("") && !tbRegOff.Text.Equals("") && !tbValue.Text.Equals("");
        }

        private void cbMode_SelectedIndexChanged(object sender, EventArgs e)
        {
            switch (cbMode.SelectedIndex)
            {
                case 0:
                    modbusMaster = new ModbusMasterASCII();
                    modbusSlave = new ModbusSlaveASCII();
                    break;
                case 1:
                    modbusMaster = new ModbusMasterRTU();
                    modbusSlave = new ModbusSlaveRTU();
                    break;
            }
            modbusSlave.SlaveID = (uint)cbSlaveID.SelectedIndex + 1;
        }

        private void tabModeControl_SelectedIndexChanged(object sender, EventArgs e)
        {
            bMaster = tabModeControl.SelectedIndex == 0;
            modbusSlave.SlaveID = (uint)cbSlaveID.SelectedIndex + 1;
        }

        private void cbSlaveID_SelectedIndexChanged(object sender, EventArgs e)
        {
            modbusSlave.SlaveID = (uint)cbSlaveID.SelectedIndex + 1;
        }

        private void bRegisters_Click(object sender, EventArgs e)
        {
            RegistersForm rf = new RegistersForm(modbusSlave);
            rf.ShowDialog();
        }

        private void cbMasterFns_SelectedIndexChanged(object sender, EventArgs e)
        {
            switch (cbMasterFns.SelectedIndex)
            {
                case 0:
                case 1:
                case 2:
                case 3:
                    lValue.Text = "Ilość rejestrów:";
                    break;
                case 4:
                case 5:
                    lValue.Text = "Wartość:";
                    break;
            }
        }
    }
}
