﻿using System;
using System.Collections.Generic;
using System.Text;

namespace Modbus
{
    abstract class ModbusSlave : Modbus
    {
        private string lastStatus;
        private uint slaveID;

        private const uint REGISTER_COUNT = 4096;

        public struct DataRow
        {
            public bool di;
            public bool coil;
            public ushort ir;
            public ushort hr;
        }

        private DataRow[] data = new DataRow[REGISTER_COUNT];

        public ModbusSlave()
        {
            Random rand = new Random(0);    // Deterministyczne
            for (int i = 0; i < REGISTER_COUNT; ++i)
            {
                data[i].di = rand.Next() % 2 != 0;
                data[i].coil = rand.Next() % 2 != 0;
                data[i].ir = (ushort)(rand.Next() % UInt16.MaxValue + 1);
                data[i].hr = (ushort)(rand.Next() % UInt16.MaxValue + 1);
            }
        }

        public uint GetDataRowCount()
        {
            return REGISTER_COUNT;
        }

        public DataRow GetDataRow(uint uIndex)
        {
            return data[uIndex];
        }

        public void SetDataRow(uint uIndex, DataRow dr)
        {
            data[uIndex] = dr;
        }

        public String LastStatus
        {
            get
            {
                return lastStatus;
            }
        }

        public uint SlaveID
        {
            get
            {
                return slaveID;
            }
            set
            {
                slaveID = value;
            }
        }

        public byte[] HandleRequest(byte[] frameData)
        {
            // Wyciaganie danych z ramki
            byte[] data = GetData(frameData);

            if(data == null || !CheckChecksum(data))
            {
                lastStatus = "Błąd transmisji, niepoprawny CRC";
                return null;
            }
            // Odrzucam zapytania do innego Slave'a
            else if (GetSlaveIDAddress(data) != 0 && GetSlaveIDAddress(data) != slaveID)
            {
                lastStatus = null;
                return null;
            }
            else
            {
                byte[] pdu = GetPDU(data);
                byte fc = pdu[0];
                ushort addr = CombineWord(pdu, 1);
                ushort regCount = CombineWord(pdu, 3);

                // Sprawdzenie funkcji
                if (!ValidateFunctionCode(fc))
                {
                    lastStatus = "Funkcja nieobsługiwana: " + fc;
                    return BuildMessage(BuildExceptionMesssage(1, fc));
                }
                // Sprawdzenie poprawności liczby rejestrów
                if (!ValidateDataValue(fc, regCount))
                {
                    lastStatus = "Nieprawidłowa ilość użytych rejestrów";
                    return BuildMessage(BuildExceptionMesssage(3, fc));
                }
                // Sprawdzanie ilości rejestrów
                if (!ValidateDataAddress(fc, addr, regCount))
                {
                    lastStatus = "Nieprawidłowa adresacja rejestrów";
                    return BuildMessage(BuildExceptionMesssage(2, fc));
                }

                byte[] output = ExecuteFunction(fc, addr, regCount);

                lastStatus = null;
                return BuildMessage(output);
            }
        }

        /* Slave functions */
        private byte[] ReadCoils(UInt16 address, UInt16 regCount)
        {
            int byteCount = (regCount - 1) / 8 + 1;
            byte[] coils = new byte[byteCount + 3];

            int iByte = 0;
            coils[iByte++] = (byte)slaveID;
            coils[iByte++] = 1;
            coils[iByte++] = (byte)byteCount;

            for (int iCoil = address; iByte < coils.Length; ++iByte)
            {
                byte mask = 0x1;

                for (int iBit = 0; iBit < 8; ++iBit)
                {
                    coils[iByte] |= data[iCoil].coil ? mask : (byte)0x0;
                    mask <<= 1;
                    ++iCoil;
                }
            }

            return coils;
        }

        private byte[] ReadDiscreteInputs(UInt16 address, UInt16 regCount)
        {
            int byteCount = (regCount - 1) / 8 + 1;
            byte[] inputs = new byte[byteCount + 3];

            int iByte = 0;
            inputs[iByte++] = (byte)slaveID;
            inputs[iByte++] = 2;
            inputs[iByte++] = (byte)byteCount;

            for (int iInput = address; iByte < inputs.Length; ++iByte)
            {
                byte mask = 0x1;

                for (int iBit = 0; iBit < 8; ++iBit)
                {
                    inputs[iByte] |= data[iInput].di ? mask : (byte)0x0;
                    mask <<= 1;
                    ++iInput;
                }
            }

            return inputs;
        }

        private byte[] ReadHoldingRegisters(UInt16 address, UInt16 regCount)
        {
            int byteCount = regCount * 2;
            byte[] hr = new byte[byteCount + 3];

            int iByte = 0;
            hr[iByte++] = (byte)slaveID;
            hr[iByte++] = 3;
            hr[iByte++] = (byte)byteCount;

            for (int iRegister = address; iByte < hr.Length; ++iRegister)
            {
                hr[iByte++] = (byte)(data[iRegister].hr >> 8);
                hr[iByte++] = (byte)(data[iRegister].hr & 0xFF);
            }

            return hr;
        }

        private byte[] ReadInputRegisters(UInt16 address, UInt16 regCount)
        {
            int byteCount = regCount * 2;
            byte[] hr = new byte[byteCount + 3];

            int iByte = 0;
            hr[iByte++] = (byte)slaveID;
            hr[iByte++] = 4;
            hr[iByte++] = (byte)byteCount;

            for (int iRegister = address; iByte < hr.Length; ++iRegister)
            {
                hr[iByte++] = (byte)(data[iRegister].ir >> 8);
                hr[iByte++] = (byte)(data[iRegister].ir & 0xFF);
            }

            return hr;
        }

        private byte[] WriteSingleCoil(UInt16 address, UInt16 value)
        {
            byte[] echo = new byte[6];

            echo[0] = (byte)slaveID;
            echo[1] = 5;
            echo[2] = (byte)(address >> 8);
            echo[3] = (byte)(address & 0xff);
            echo[4] = (byte)(value >> 8);
            echo[5] = (byte)(value & 0xff);

            data[address].coil = (value == 0xFF00);

            return echo;
        }

        private byte[] WriteSingleRegister(UInt16 address, UInt16 value)
        {
            byte[] echo = new byte[6];

            echo[0] = (byte)slaveID;
            echo[1] = 6;
            echo[2] = (byte)(address >> 8);
            echo[3] = (byte)(address & 0xff);
            echo[4] = (byte)(value >> 8);
            echo[5] = (byte)(value & 0xff);

            data[address].hr = value;

            return echo;
        }

        /* Validate and common functions */
        private byte[] ExecuteFunction(byte fun, ushort addr, ushort value)
        {
            switch (fun)
            {
                case 1:
                    return ReadCoils(addr, value);
                case 2:
                    return ReadDiscreteInputs(addr, value);
                case 3:
                    return ReadHoldingRegisters(addr, value);
                case 4:
                    return ReadInputRegisters(addr, value);
                case 5:
                    return WriteSingleCoil(addr, value);
                case 6:
                    return WriteSingleRegister(addr, value);
            }
            return null;
        }

        private bool ValidateFunctionCode(byte fun)
        {
            switch (fun)
            {
                case 1:
                case 2:
                case 3:
                case 4:
                case 5:
                case 6:
                    return true;
            }
            return false;
        }

        private bool ValidateDataValue(byte fun, UInt16 value)
        {
            switch (fun)
            {
                case 1:
                case 2:
                    return value > 0 && value <= 0x07D0;
                case 3:
                case 4:
                    return value > 0 && value <= 0x007D;
                case 5:
                    return value == 0xFF00 || value == 0x0000;
                case 6:
                    return true;
            }
            return false;
        }
        
        private bool ValidateDataAddress(byte fun, UInt16 address, UInt16 value)
        {
            switch (fun)
            {
                case 1:
                case 2:
                case 3:
                case 4:
                    return address >= 0 && address <= REGISTER_COUNT && (address + value - 1) >= 0 && (address + value - 1) <= REGISTER_COUNT;
                case 5:
                case 6:
                    return address >= 0 && address <= REGISTER_COUNT;
            }
            return false;
        }

        private UInt16 CombineWord(byte[] data, uint offset)
        {
            return (ushort)(data[offset] << 8 | data[offset + 1]);
        }

        private byte[] BuildExceptionMesssage(byte ec, byte fun)
        {
            byte[] data = new byte[2];
            data[0] = (byte)(fun + 0x80);
            data[1] = ec;
            return data;
        }

        protected uint GetSlaveIDAddress(byte[] data)
        {
            return data[0];
        }

        protected byte[] GetPDU(byte[] data)
        {
            byte[] pdu = new byte[data.Length - 2];

            for (uint i = 1; i < data.Length - 1; ++i)
            {
                pdu[i - 1] = data[i];
            }

            return pdu;
        }

        abstract protected byte[] GetData(byte[] data);

        abstract protected bool CheckChecksum(byte[] data);
    }
}
